@@ -1,25 +1,10 @@
 /*
-
     TiMidity -- Experimental MIDI to WAVE converter
     Copyright (C) 1995 Tuukka Toivonen <toivonen@clinet.fi>
 
     This program is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-
-    playmidi.c -- random stuff in need of rearrangement
-
-*/
+    it under the terms of the Perl Artistic License, available in COPYING.
+ */
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -1691,34 +1676,7 @@ void Timidity_SetVolume(int volume)
   ctl->master_volume(amplification);
 }
 
-MidiSong *Timidity_LoadSong(const char *midifile)
-{
-  MidiSong *song;
-  int32 events;
-  SDL_RWops *rw;
-
-  /* Allocate memory for the song */
-  song = (MidiSong *)safe_malloc(sizeof(*song));
-  memset(song, 0, sizeof(*song));
-
-  /* Open the file */
-  strcpy(midi_name, midifile);
-
-  rw = SDL_RWFromFile(midifile, "rb");
-  if ( rw != NULL ) {
-    song->events=read_midi_file(rw, &events, &song->samples);
-    SDL_RWclose(rw);
-  }
-
-  /* Make sure everything is okay */
-  if (!song->events) {
-    free(song);
-    song = NULL;
-  }
-  return(song);
-}
-
-MidiSong *Timidity_LoadSong_RW(SDL_RWops *rw)
+MidiSong *Timidity_LoadSong_RW(SDL_RWops *rw, int freerw)
 {
   MidiSong *song;
   int32 events;
@@ -1729,7 +1687,10 @@ MidiSong *Timidity_LoadSong_RW(SDL_RWops *rw)
 
   strcpy(midi_name, "SDLrwops source");
 
-  song->events=read_midi_file(rw, &events, &song->samples);
+  song->events = read_midi_file(rw, &events, &song->samples);
+  if (freerw) {
+    SDL_RWclose(rw);
+  }
 
   /* Make sure everything is okay */
   if (!song->events) {
